# Program for a GA
# This genetic algorithm is partially based on the algorithm 3 proposed by Coley, David A.
# An Introduction to Genetic Algorithms for Scientists and Engineers.
# Singapore: World Scientific Publishing Co. Pte. Ltd., 1999. 227 pp.
# See page 27 of the above reference.
# In this version, the selection operator uses Roulette wheel and elitsm is present.
#
# Dr. Walter F. de Azevedo Jr.
# azevedolab.net
# Porto Alegre, April 25th 2017
#

class Simple_GA02(object):
    """A class to implement a simple GA"""
    
    # Constructor method
    def __init__(self,pop_size,max_iter,p_cross,len_str,p_mut,f_min,f_max,conv_f,seed_in):
                
        # Define atributes
        self.pop_size = pop_size            # Population size
        self.max_iter = max_iter            # Number of iterations
        self.p_cross = p_cross              # Probability of crossover
        self.len_str = len_str              # String length
        self.p_mut = p_mut                  # Probability of mutation_Coley_algo_03
        self.f_min = f_min                  # Minimum float
        self.f_max = f_max                  # Maximum float
        self.conv_f = conv_f                # Conversion factor (fitness function to dimensions)   
        self.seed_in = seed_in              # Seed for generation of random numbers
            
    # __str__ method
    def __str__(self):
        info = "Simple_GA02 class\n"
        info += "Class to generate a model based on a simple genetic algorithm.\n"
        info += "It is an OOP implementation of a genetic algorithm partially based on\n"
        info += "the algorithm 3 from Coley, 1999.\n"
        info += "ReferenceÇ\n"
        info += " Coley, David A. An Introduction to Genetic Algorithms for Scientists and Engineers.\n"
        info += "Singapore: World Scientific Publishing Co. Pte. Ltd., 1999. 227 pp."
        
        return info
    
    # Definition of get_max()
    def get_max(self,x):
        """Method to return the maximum value for an array"""
    
        # Import library
        import numpy as np
    
        # Find maximum
        max1 = np.max(x)
    
        return max1

    # Definition of get_mean() f
    def get_mean(self,x):
        """Method to return the mean value for an array"""
    
        # Import library
        import numpy as np
    
        # Find mean value
        mean1 = np.mean(x)
    
        return mean1
            
    # Definition of gen_bin_string_seed() 
    def gen_bin_string_seed(self):
        """Method to generate binary strings"""
    
        # Import library
        import numpy as np
        
        # Set up seed for random number
        np.random.seed(self.seed_in)
        
        # Set up arrays of zeros
        self.max_array = np.zeros(self.max_iter)
        self.mean_array = np.zeros(self.max_iter)
    
        # Set up empty list
        bin_str_list = []
            
        # Generate binary string
        for i in range(self.pop_size):
            bin1 = np.random.randint(2,size=self.len_str)    # Binary (2) strings of length of len_str
            bin_str0 = str(bin1)
            bin_str0 = bin_str0.replace(" ","")
            bin_str0 = bin_str0.replace("[","")
            bin_str0 = bin_str0.replace("]","")
            bin_str_list.append(bin_str0)
        
        # Update population
        self.current_pop = bin_str_list
        
        print("\nInitial Population")
        # Looping through initial population
        for bin_string in self.current_pop:
            print(bin_string)
            
    # Definition of bin2dec()
    def bin2dec(self,string_num):
        """Method to convert binary string to decimal integer"""
    
        # Convert binary to decinal
        num1 = str(int(string_num,2))
    
        return num1

    # Definition of bin2dec_array() 
    def bin2dec_array(self,my_pop_in):
        """Method to convert a list of binary strings to an array of integer decimals"""
    
        # Import libray
        import numpy as np
    
        # Set up an array of zeros with self.pop_size elements
        x = np.zeros(self.pop_size)
    
        # Looping through population to convert from binary to an array of decimals 
        i = 0
        for line in my_pop_in:
            aux = self.bin2dec(line)
            x[i] = aux
            i+=1
    
        return x
    
    # Definition of shuffle()
    def shuffle(self,pop_in):
        """Method to shuffle population"""
    
        # Import library
        import numpy as np
                    
        # Shuffle by calling random.shuffle()
        np.random.shuffle(pop_in)
        
        return pop_in
            
    # Definition of gen_offspring method
    def gen_offsprings(self,str1,str2):
        """Method to generate offsprings"""
    
        # Import library
        import numpy as np
            
        # Define partition locus
        locus = np.random.randint(2,len(str1)-1)
    
        # Generate offspring
        offspring1 = str1[0:locus] + str2[locus:]
        offspring2 = str2[0:locus] + str1[locus:]
    
        return offspring1, offspring2
    
    # Definition of single_point_crossover
    def single_point_crossover(self,new_pop_in):
        """Method to update population after crossover"""
        
        # Import library
        import numpy as np
                
        # Assign population to count_pop. This population size is less than self.pop_size
        count_pop = 0
        
        # Loop while to generate new population with offspring
        while count_pop < self.pop_size:    # To make sure the size of population stay the same
            
            # Get random number
            r_n = np.random.uniform(0,1)
            
            # Test cross over prob 
            if float(self.p_cross) < r_n:                   
                
                # Loop for to generate offspring (gen_offsprings())
                for i in range(int(self.pop_size/2)):
                    o1,o2 = self.gen_offsprings(new_pop_in[i],new_pop_in[i+1])
                    new_pop_in.append(o1)
                    new_pop_in.append(o2)
                    count_pop += 2
            
            # Get random number
            r_n = np.random.uniform(0,1)
        
        # Assign new population to self.current_pop
        self.current_pop = new_pop_in
    
    # Definition of int2float()
    def int2float(self,i):
        """Method to convert integer to float"""
        
        f = self.f_min + i*(self.f_max - self.f_min)/(2**(self.len_str) -1)
        
        return f
        
    # Definition of roulette wheel as shown in the Algorithm 2 from Coley 1999
    def roulette_wheel(self):
        """Method for application of roulette wheel to a population"""
        
        # Import library
        import numpy as np
                        
        # Sum fitness function vaules
        fit_sum = np.sum(self.fit)
        
        # Set up empty list
        new_pop = []
        
        # Looping through all individuals
        for j in range(self.pop_size):
            
            # Generate random number
            rs =np.random.uniform(0,fit_sum)
            
            # Set up initial value for fit_partial
            fit_partial = 0
            
            # Looping through inidividual until fit_partial > rs
            for i in range(self.pop_size):
                fit_partial += self.fit[i]
                
                # Check if condition is satisfied
                if fit_partial > rs:
                    new_pop.append(self.current_pop[i])
                    break
        
        # Update population
        self.current_pop = new_pop
        
    # Definition of elitism method
    def elitism(self):
        """Method for elitism"""
        
        # Import libraries
        import numpy as np
        import test_function as tf
        
        # Call bin2dec_array()
        dec = self.bin2dec_array(self.current_pop)
                        
        # Call int2float()
        my_float = self.int2float(dec)
        
        # Instantiating an object of the Function() class and assigns it to the variable fit0
        fit0 = tf.Function(my_float)
        
        # Invoking the fitness_funct() method 
        my_fit = fit0.fitness_func()
        
        # Set up an array
        my_array = np.array(my_fit)
                
        # Find the maximum for the fitness function and assigns it to max_fit
        max_fit = np.max(my_array)
        
        # Set up an empty list
        new_pop = []
        
        # Find the position for the max_fit in the my_array
        pos_array = np.where( my_array == max_fit)
        
        # Get the integer for index
        my_index = int(pos_array[0][0])
    
        # Assign elite member to self.elite
        self.elite = self.current_pop[my_index]
        
        # Show elite member
        print("Elite member: ", self.elite, max_fit)
    
    # Include elite member in the population
    def include_elite(self):
        """Method to include elite member in the current population"""
        
        # Import library
        import numpy as np
                
        # Generate random number
        rs = np.random.randint(0,self.pop_size)
        
        # Set up empty list
        new_pop = []
        
        # Looping through population and change for elite meber at random position
        for i in range(self.pop_size):
            if i != rs:
                new_pop.append(self.current_pop[i])
            else:
                new_pop.append(self.elite)
        
        # Update current_pop
        self.current_pop = new_pop
    
    # Definition of selection_Coley_algo_03() (selection operator)
    def selection_Coley_algo_03(self):
        """Method for selection operator using aklgorithm 2 from Coley 1999"""
        
        # Import libray
        import test_function as tf
        
        # Call bin2dec_array()
        dec = self.bin2dec_array(self.current_pop)
                        
        # Call int2float()
        my_float = self.int2float(dec)
    
        # Instantiating an object of the Function() class and assigns it to the variable fit1
        fit1 = tf.Function(my_float)
        
        # Invoking the fitness_funct() method 
        self.fit = fit1.fitness_func()
        
        # Call elitism
        self.elitism()
                
        # Call roulette_wheel()
        self.roulette_wheel()      
            
    # Definition of cross_over 
    def crossover_Coley_algo_03(self):
        """Method to apply crossover operator to a population"""
    
        # Call shuffle()
        self.current_pop = self.shuffle(self.current_pop)
        
        # Call single_point_crossover
        self.single_point_crossover(self.current_pop)

    # Definition of mutation_Coley_algo_03 method
    def mutation_Coley_algo_03(self):
        """Method to apply mutation_Coley_algo_03 operator"""
        
        # Determine total
        total = self.pop_size*self.len_str
    
        # Some editing list-> string
        str_pop = "".join(self.current_pop)
        str_pop = str_pop.replace("\n","")
    
        # Set up empty string
        str_pop_out = ""
    
        # Carry out mutation
        for i in range(total):
            if i == int(1/self.p_mut):
                if str_pop[i] == 1:
                    str_pop_out+="0"
                else:
                    str_pop_out+="1"
            else:
                str_pop_out+=str_pop[i]
    
        # Back to string. Set up empty list
        pop = []
    
        # Loop to recover population after mutation_Coley_algo_03
        for i in range(self.pop_size):
            aux = ""
            for j in range(self.len_str):
                aux += str_pop_out[i+j]
            pop.append(aux)
        
        self.current_pop = pop
    
    # Definition of show_pop()
    def show_pop(self,string2show,pop2show,f_in,function_in):
        """Method to show population"""
        
        # Show information concerning the population
        print("\n"+string2show)
        
        # Looping though the elements of the population
        for i in range(len(pop2show)):
            print(pop2show[i],"  %.3f"%f_in[i],"  %.3f"%function_in[i])
        
    # Definition of plot_ga_evolution() method
    def plot_ga_evolution(self,x_label,y_label,legend1,legend2,file_out):
        """Method to generate a plot of the evolution of the fitness function values for each iteraction"""
    
        # Import libraries
        import numpy as np
        import matplotlib.pyplot as plt
    
        # Generate array for x-axis
        x = np.linspace(1,self.max_iter,self.max_iter)
            
        # Generate plot
        plt.plot(x,self.max_array,label=legend1)
        plt.plot(x,self.mean_array,label=legend2)
        plt.legend(loc = "lower right")
        plt.xlabel(x_label)
        plt.ylabel(y_label)
        plt.grid(True)
        plt.show()
        plt.savefig(file_out)
    
    # Definition of coley_algo_03 method
    def coley_algo_03(self):
        """Method to implement a genetic algorithm partially based on the algorithm 3 from Coley 1999"""
        
        # Import libraries 
        import numpy as np
        import touche as tart
        import hexadec as hx
        import test_function as tf
        
        # General parameters for the turtles
        x_min = -240        # Minimum x coordinate in pixels
        x_max =  240        # Maximum x coordinate in pixels
        y_min = -240        # Minimum y coordinate in pixels
        y_max =  240        # Maximum y coordinate in pixels
        max_rotate = 8      # Maximum rotation angle 8*45 degrees
                    
        # Call gen_bin_string_seed()
        self.gen_bin_string_seed()
                
        # Looping through the number of iteractions
        for i in range(self.max_iter):
            
            # Call selection_Coley_algo_03()
            self.selection_Coley_algo_03()
            
            # Call crossover_Coley_algo_03() method
            self.crossover_Coley_algo_03()
            
            # Call mutation_Coley_algo_03 method
            self.mutation_Coley_algo_03()
            
            # Call include_elite
            self.include_elite()
            
            # Call bin2dec_array()
            dec = self.bin2dec_array(self.current_pop)
                        
            # Call int2float()
            my_float = self.int2float(dec)
                                    
            # Instantiating an object of the Function() class and assigns it to the variable fit2
            fit2 = tf.Function(my_float)
            
            # Invoking the fitness_funct() method 
            fitness = fit2.fitness_func()
            
            # Call show_pop()
            self.show_pop("\nPopulation for generation "+str(i+1),self.current_pop,my_float,fitness)
            
            # Instantiating an object of the Hexa() class and assigns it to the variable h1 
            h1 = hx.Hexa()                              # String for hexadecimal
            
            # Invoking the method gen_hex()
            h1.gen_hex()
            hex = h1.my_hex_string
            
            # Looping through all elements
            for j in range(self.pop_size):
                
                # Define parameters for each turtle
                x = np.random.randint(x_min,x_max)          # x coordinate in pixels
                y = np.random.randint(y_min,y_max)          # y coordinate in pixels
                length  = fitness[j]*self.conv_f            # Length for the turtle
                width = fitness[j]*self.conv_f              # Width for the turtle
                rotate = np.random.randint(0,max_rotate)    # Rotation angle 8*45 degrees

                # Instantiating an object of the Touche() class and assigns it to the variable t 
                t = tart.Touche(x, y, length, width, rotate, "#"+hex) 

                # Invoking the method draw()
                t.draw()
                        
            # Call get_max()
            self.max_array[i] = self.get_max(fitness)
            
            # Call get_mean()
            self.mean_array[i] = self.get_mean(fitness)
        
        # Invoking the method bye()   
        t.bye()
    
        # Call show_pop()
        self.show_pop("\nFinal Population",self.current_pop,my_float,fitness)
        
def main():  

    # Set up initial values for GA
    # pop_size    = 8            # Population size
    # max_iter    = 100           # Number of iterations
    # p_cross     = 0.8          # Probability of crossover
    # len_str     = 12           # Length of strings
    # p_mut       = 0.3         # Probability of mutation_Coley_algo_03
    # seed        = 314159       # Seed for generation of random number
    # min_f       = 0.0          # Minimum float
    # max_f       = 100.0        # Maximum float
    # conv_factor = 0.0005       # Conversion factor (fitness function to dimensions)

    # Import read_ga.py
    import read_ga as rga

    # Call read_GA_parameters()
    pop_size,max_iter,p_cross,len_str,p_mut,seed,min_f,max_f,conv_factor = rga.read_GA_parameters()

    # Instantiating an object of the Simple_GA02() class and assigns it to the variable p1 
    p1 = Simple_GA02(pop_size,max_iter,p_cross,len_str,p_mut,min_f,max_f,conv_factor,seed)
    
    # Invoking the plot_ga_evolution() method 
    p1.coley_algo_03()
    
    # Invoking the coley_algo_03() method
    p1.plot_ga_evolution("Generation","Fitness function, f","Maximum value of f","Mean value of f","ga01.png")
        
main()
